/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 21/09/2007
   derniere modification : 18/07/2008 */

#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEList.h>


/* ajoute le 21/09/2007 */
/* initialise un iterateur */
static void SCE_List_InitIt (SCE_SListIterator *it)
{
    it->next = it->prev = NULL;
    it->data = NULL;
}

/* ajoute le 23/09/2007 */
static SCE_SListIterator* SCE_List_CreateIt (void)
{
    SCE_SListIterator *it = SCE_malloc (sizeof *it);
    if (!it)
        Logger_LogSrc ();
    else
        SCE_List_InitIt (it);
    return it;
}

/* ajoute le 23/09/2007 */
/* revise le 11/01/2008 */
static void SCE_List_DeleteIt (SCE_SListIterator *it)
{
    SCE_free (it);
}


/* ajoute le 21/09/2007 */
/* revise le 02/10/2007 */
/**
 * \brief Initialize a SCE_SList
 * \param l the SCE_SList to initialize
 * 
 * This function initializes a SCE_SList structure.
 */
void SCE_List_Init (SCE_SList *l)
{
    l->first = NULL;
    l->f = NULL;
}


/* ajoute le 21/09/2007 */
/* revise le 22/09/2007 */
/**
 * \brief Creates a SCE_SList
 * \param f a SCE_FListFreeFunc function to free the list's data at deletion
 * \returns a newly allocated SCE_SList, or NULL on error
 * \sa SCE_List_SetFreeFunc()
 * 
 * This function creates a new SCE_SList structure and initialize it; then you
 * haven't to init it yourself.
 * The SCE_FListFreeFunc is used to free the list's data when calling a free
 * function like SCE_List_Clear() or SCE_List_Delete(). If it is NULL, the
 * list's data will not be freed.
 */
SCE_SList* SCE_List_Create (SCE_FListFreeFunc f)
{
    SCE_SList *l = NULL;
    if (!(l = SCE_malloc (sizeof *l)))
        Logger_LogSrc ();
    else
    {
        SCE_List_Init (l);
        l->f = f;
    }
    return l;
}

/* ajoute le 24/09/2007 */
/**
 * \brief Clear a SCE_SList
 * \param l the SCE_SList to clear
 * \sa SCE_List_Delete()
 * \sa SCE_List_Create()
 * 
 * This function clears any data in a SCE_SList by using the registered free
 * function (see SCE_List_Create()); but leave the list istelf reusable without
 * any reallocation.
 * Use this function if you would change all data in the list.
 * If you just want to free the list, use SCE_List_Delete() instead.
 */
void SCE_List_Clear (SCE_SList *l)
{
    SCE_SListIterator *it = l->first;
    SCE_SListIterator *tmp = NULL;

    while (it)
    {
        tmp = it;
        it = it->next;
        if (l->f)
            l->f (tmp->data);
        SCE_List_DeleteIt (tmp);
    }

    l->first = NULL;
}

/* ajoute le 21/09/2007 */
/* revise le 24/09/2007 */
/**
 * \brief Delete a SCE_SList
 * \param l the SCE_SList to delete
 * \sa SCE_List_Clear()
 * 
 * This function deletes a SCE_SList and all it's data by calling
 * SCE_List_Clear().
 * After a call of this function, the list is fully deleted and can't be reused.
 */
void SCE_List_Delete (SCE_SList *l)
{
    if (l)
    {
        SCE_List_Clear (l);
        SCE_free (l);
    }
}


/* ajoute le 22/09/2007 */
/* revise le 18/07/2008 */
/**
 * \brief Prepends data to a SCE_SList
 * \param l the SCE_SList where data has to be prepended
 * \param d the data to prepend to the list
 * \returns SCE_OK on success, or SCE_ERROR on failure
 * 
 * This function prepends data to a SCE_SList.
 */
int SCE_List_AddFront (SCE_SList *l, void *d)
{
    if (!l->first)
    {
        l->first = SCE_List_CreateIt ();
        if (!l->first)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        l->first->data = d;
    }
    else
    {
        if (SCE_List_AddFrontIt (l->first, d) < 0)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        l->first = l->first->prev;
    }
    return SCE_OK;
}
/**
 * \brief Appends data to a SCE_SList
 * \param l the SCE_SList where data has to be appended
 * \param d the data to append to the list
 * \returns SCE_OK on success, or SCE_ERROR on failure
 * 
 * This function appends data to a SCE_SList.
 */
int SCE_List_AddBack (SCE_SList *l, void *d)
{
    if (!l->first)
    {
        l->first = SCE_List_CreateIt ();
        if (!l->first)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        l->first->data = d;
    }
    else if (SCE_List_AddBackIt (SCE_List_GetLast (l), d) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 07/05/2008 */
/* revise le 18/07/2008 */
/**
 * \brief Prepends data to a list iterator
 * \param i the SCE_SListIterator where data has to be prepended
 * \param d the data to prepend to the iterator
 * \returns SCE_OK on success, or SCE_ERROR on failure
 * 
 * This function prepends data to a SCE_SListIterator.
 */
int SCE_List_AddFrontIt (SCE_SListIterator *i, void *d)
{
    if (i)
    {
        SCE_SListIterator *new = SCE_List_CreateIt ();
        if (!new)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        new->data = d;
        SCE_List_InsertFront (i, new);
    }
    return SCE_OK;
}
/* ajoute le 07/05/2008 */
/* revise le 18/07/2008 */
/**
 * \brief Appends data to a list iterator
 * \param i the SCE_SListIterator where data has to be appended
 * \param d the data to append to the iterator
 * \returns SCE_OK on success, or SCE_ERROR on failure
 * 
 * This function appends data to a SCE_SListIterator.
 */
int SCE_List_AddBackIt (SCE_SListIterator *i, void *d)
{
    if (i)
    {
        SCE_SListIterator *new = SCE_List_CreateIt ();
        if (!new)
        {
            Logger_LogSrc ();
            return SCE_ERROR;
        }
        new->data = d;
        SCE_List_InsertBack (i, new);
    }
    return SCE_OK;
}


/* ajoute le 18/07/2008 */
void SCE_List_InsertFront (SCE_SListIterator *it, SCE_SListIterator *new)
{
    if (it->prev)
        it->prev->next = new;
    new->prev = it->prev;
    new->next = it;
    it->prev = new;
}
/* ajoute le 18/07/2008 */
void SCE_List_InsertBack (SCE_SListIterator *it, SCE_SListIterator *new)
{
    if (it->next)
        it->next->prev = new;
    new->next = it->next;
    new->prev = it;
    it->next = new;
}


/* ajoute le 22/09/2007 */
/* revise le 11/01/2008 */
/**
 * \brief Remove the first element of a list
 * \param l the SCE_SList from where detach data
 * \returns the data of the removed element
 * \see SCE_List_EraseFront()
 * 
 * This function removes the first element from the given list and returns its
 * data. If you just want to remove the first element and its data, use
 * SCE_List_EraseFront().
 */
void* SCE_List_RemoveFront (SCE_SList *l)
{
    void *d = NULL;
    SCE_SListIterator *it = l->first;

    if (it)
    {
        if (it->next)
            it->next->prev = NULL;

        l->first = it->next;
        d = it->data;
        SCE_List_DeleteIt (it);
    }

    return d;
}


/* ajoute le 22/09/2007 */
/* revise le 08/10/2007 */
/**
 * \brief Remove the last element of a list
 * \param l the SCE_SList from where detach data
 * \returns the data of the removed element
 * \see SCE_List_EraseBack()
 * \see SCE_List_RemoveFront()
 * 
 * See documentation of SCE_List_RemoveFront().
 */
void* SCE_List_RemoveBack (SCE_SList *l)
{
    void *d = NULL;
    SCE_SListIterator *it = SCE_List_GetLast (l);

    if (it)
    {
        if (it->prev)
            it->prev->next = NULL;
        else
            l->first = NULL;
        d = it->data;
        SCE_List_DeleteIt (it);
    }

    return d;
}

/* ajoute le 22/09/2007 */
/* revise le 01/10/2007 */
/**
 * \brief Remove an element of a list
 * \param l the SCE_SList from where detach data
 * \param it the iterator of the element to be removed
 * \returns the data of the removed element
 * \see SCE_List_Erase()
 * 
 * This function removes the element represented by the iterator from the given
 * list and returns its data.
 * If you just want to remove an element and its data, use SCE_List_Erase().
 */
void* SCE_List_Remove (SCE_SList *l, SCE_SListIterator *it)
{
    void *d = it->data;

    if (it->prev)
        it->prev->next = it->next;
    else
        l->first = it->next;

    if (it->next)
        it->next->prev = it->prev;

    SCE_List_DeleteIt (it);

    return d;
}



/* ajoute le 22/09/2007 */
/* revise le 25/09/2007 */
/**
 * \brief Fully remove the first element of a list
 * \param l the SCE_SList from where detach data
 * 
 * This function removes the first element and its data from the given list.
 */
void SCE_List_EraseFront (SCE_SList *l)
{
    void *d = SCE_List_RemoveFront (l);
    if (l->f)
        l->f (d);
}

/* ajoute le 22/09/2007 */
/* revise le 25/09/2007 */
/**
 * \brief Fully remove the last element of a list
 * \param l the SCE_SList from where detach data
 * \sa SCE_List_EraseFront()
 */
void SCE_List_EraseBack (SCE_SList *l)
{
    void *d = SCE_List_RemoveBack (l);
    if (l->f)
        l->f (d);
}

/* ajoute le 22/09/2007 */
/* revise le 01/10/2007 */
/**
 * \brief Fully remove an element of a list
 * \param l the SCE_SList from where remove element
 * \param it the iterator of the element to be removed
 * 
 * This function removes the element represented by the iterator from the given
 * list. This function deletes the data of the element, ulike SCE_List_Remove()
 * that remove the element, and return the data.
 */
void SCE_List_Erase (SCE_SList *l, SCE_SListIterator *it)
{
    void *d = SCE_List_Remove (l, it);
    if (l->f)
        l->f (d);
}

/* ajoute le 26/03/2008 */
/**
 * \brief Sets data of an iterator
 * \param it the SCE_SListIterator where set the data
 * \param data the data to set to the iterator
 * \returns the old data, if any
 */
void* SCE_List_SetData (SCE_SListIterator *it, void *data)
{
    void *old = it->data;
    it->data = data;
    return old;
}
/* ajoute le 24/09/2007 */
/**
 * \brief Gets data of an iterator
 * \param it the SCE_SListIterator from where get data
 * \returns the data, if any
 */
void* SCE_List_GetData (SCE_SListIterator *it)
{
    return it->data;
}


/**
 * \brief Gets the length of a SCE_SList
 * \param l a SCE_SList
 * \returns the length of the list pointed by \p l
 */
unsigned int SCE_List_GetSize (SCE_SList *l)
{
    unsigned int n = 0;
    SCE_SListIterator *it;
    SCE_List_ForEach (it, l)
        n++;
    return n;
}

/* ajoute le 24/04/2008 */
/**
 * \brief Gets the first iterator of a list
 * \param l the list from where get the iterator
 * \returns the first iterator of the list, if any
 */
SCE_SListIterator* SCE_List_GetFirst (SCE_SList *l)
{
    return l->first;
}
/* ajoute le 22/09/2007 */
/* revise le 23/09/2007 */
/**
 * \brief Gets the last iterator of a list
 * \param l the list from where get the iterator
 * \returns the last iterator of the list, if any
 */
SCE_SListIterator* SCE_List_GetLast (SCE_SList *l)
{
    SCE_SListIterator *it = l->first;
    if (it)
    {
        while (it->next)
            it = it->next;
    }
    return it;
}

/* ajoute le 26/03/2008 */
/**
 * \brief Gets the next iterator of an interator
 * \param it the base iterator
 * \returns the next iterator, if any
 */
SCE_SListIterator* SCE_List_GetNext (SCE_SListIterator *it)
{
    return it->next;
}
/**
 * \brief Gets the previous iterator of an interator
 * \param it the base iterator
 * \returns the previous iterator, if any
 */
SCE_SListIterator* SCE_List_GetPrev (SCE_SListIterator *it)
{
    return it->prev;
}

/* ajoute le 26/09/2007 */
/* revise le 08/10/2007 */
/**
 * \brief Gets the index position of an iterator
 * \param it the iterator from which you would know the index
 * \returns the index position of the iterator
 */
unsigned int SCE_List_GetIndex (SCE_SListIterator *it)
{
    unsigned int n = 0;
    SCE_List_ForEachPrev (it)
        n++;
    return n - 1;
}

/* ajoute le 26/09/2007 */
/**
 * \brief Gets an iterator of a list from its index position
 * \param l the SCE_SList where search for iterator
 * \param n the index of the iterator you're searching for
 * \returns the iterator at \p n position, if any
 */
SCE_SListIterator* SCE_List_GetIterator (SCE_SList *l, unsigned int n)
{
    SCE_SListIterator *it = NULL;
    unsigned int i = 0;
    
    SCE_List_ForEach (it, l)
    {
        if (i == n)
            return it;
        i++;
    }
    /* iterateur numero 'n' non trouve dans la liste */
    Logger_Log (/*SCE_STACK_OVERFLOW*/SCE_INVALID_ARG);
    return NULL;
}

/* ajoute le 15/10/2007 */
/**
 * \brief Gets an iterator of a list by its data
 * \param l an SCE_SList list
 * \param data the data to compare to each list's data
 * \param f a function to compares \p data to each list's data, or NULL for a
 * pointer comparison
 * \returns the founded iterator, if any
 * 
 * This function searches for an iterator containing \p data.
 * The comparison is done by calling \p f and/or comparing the pointers.
 */
SCE_SListIterator* SCE_List_LocateIterator (SCE_SList *l, void *data,
                                            SCE_FListCompareData f)
{
    SCE_SListIterator *i = NULL;
    SCE_List_ForEach (i, l)
    {
        if ((f && f (data, i->data)) || data == i->data)
            return i;
    }
    return NULL;
}

/* ajoute le 20/10/2007 */
/**
 * \brief Gets an iterator's index of a list by its data
 * \param l an SCE_SList list
 * \param data the data to compare to each list's data
 * \param f a function to compares \p data to each list's data, or NULL for a
 * pointer comparison
 * \returns the founded iterator's index, or 0 if index can't be found
 * \sa SCE_List_LocateIterator()
 */
unsigned int SCE_List_LocateIndex (SCE_SList *l, void *data,
                                   SCE_FListCompareData f)
{
    unsigned int j = 0;
    SCE_SListIterator *i = NULL;
    SCE_List_ForEach (i, l)
    {
        if ((f && f (data, i->data)) || data == i->data)
            return j;
        j++;
    }
    return 0;
}

/* ajoute le 25/09/2007 */
/**
 * \brief Sets the SCE_FListFreeFunc of a list
 * \param l an SCE_SList
 * \param f the SCE_FListFreeFunc function to set as data's free function
 * \sa SCE_FListFreeFunc
 * 
 * This function sets the function called to free user's data in functions tha
 * free all data, like SCE_List_Erase()
 */
void SCE_List_SetFreeFunc (SCE_SList *l, SCE_FListFreeFunc f)
{
    l->f = f;
}
