/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 22 decembre 2006
   derniere modification le 22/08/2008 */

#include <stdlib.h>
#include <string.h>
#include <GL/glew.h> /* pour les types GL... */

#include <SCE/utils/SCEError.h>
#include <SCE/utils/SCEMemory.h>

/**
 * \file SCEMemory.c
 * \copydoc memory
 * \brief Memory managment
 * 
 * \file SCEMemory.h
 * \copydoc memory
 * \brief Memory managment
 */

/**
 * \defgroup memory Memory managment
 * Wrapper around libc functions for memory managment
 * \ingroup utils
 */

/** @{ */


/**
 * \brief Stores metadata about memory blocks
 *
 * Used to know where a bloc was allocated.
 */
typedef struct SCE_MemAlloc
{
    const char *file;
    unsigned int line;
    size_t size;
    void *ptr;

    struct SCE_MemAlloc *next, *prev;
} SCE_MemAlloc;


/*! Table of all allocations */
static SCE_MemAlloc allocs = {"root allocations list", 0, 0, NULL, NULL, NULL};

#define SCE_Mem_For(i) for (i = allocs.next; i; i = i->next)

/* fonction pour gerer la table des allocations */
static void SCE_Mem_InitAlloc (SCE_MemAlloc *m)
{
    m->file = NULL;
    m->line = 1;
    m->size = 0;
    m->ptr = NULL;
    m->next = m->prev = NULL;
}

static SCE_MemAlloc* SCE_Mem_NewAlloc (void)
{
    SCE_MemAlloc *m = malloc (sizeof *m);
    if (m)
        SCE_Mem_InitAlloc (m);
    else
        Logger_Log (SCE_OUT_OF_MEMORY);
    return m;
}

static void SCE_Mem_DeleteAlloc (SCE_MemAlloc *m)
{
    if (m)
    {
        free (m->ptr);
        free (m);
    }
}


static SCE_MemAlloc* SCE_Mem_LocateAllocFromPointer (void *p)
{
    SCE_MemAlloc *i = &allocs;

    while (i->next && i->ptr != p)
        i = i->next;

    if (i->ptr == p)
        return i;

    return NULL;
}


int SCE_Mem_IsValid (void *p)
{
    return (SCE_Mem_LocateAllocFromPointer (p) != NULL);
}


static void SCE_Mem_AddAlloc (SCE_MemAlloc *m)
{
    SCE_MemAlloc *i = &allocs;

    while (i->next)
        i = i->next;

    i->next = m;
    m->prev = i;
}

static int SCE_Mem_RemoveAlloc (void *p)
{
    SCE_MemAlloc *i = &allocs;

    while (i->next && i->ptr != p)
        i = i->next;

    if (i->ptr == p)
    {
        /* element located */
        i->prev->next = i->next;
        if (i->next)
            i->next->prev = i->prev;
        SCE_Mem_DeleteAlloc (i);
        return 0;
    }

    return -1;
}


/**
 * \brief SCE's malloc wrapper
 * \param s Size wanted for the block
 * \param file, line Where the block is asked
 * \returns A pointer to a newly allocated block on succes, NULL on failure
 * 
 * You will generally want to call SCE_malloc() that wraps this function.
 * 
 * \see SCE_malloc()
 */
void* SCE_Mem_Alloc (const char *file, unsigned int line, size_t s)
{
    void *p = NULL;
    SCE_MemAlloc *mem = NULL;

    mem = SCE_Mem_NewAlloc ();
    if (!mem) /* pas de LogSrc, car cette fonction alloue de la memoire */
        return NULL;

    p = malloc (s);
    if (!p)
    {
        SCE_Mem_DeleteAlloc (mem);
        Logger_Log (SCE_OUT_OF_MEMORY);
        return NULL;
    }

    mem->file = file;
    mem->line = line;
    mem->size = s;
    mem->ptr = p;

    SCE_Mem_AddAlloc (mem);

    return p;
}

/**
 * \brief Callocs wrapper
 * \param s Size of one item
 * \param n Num of item required
 * \param file, line Where the block is asked
 * \returns A pointer to a newly allocated block on success, NULL on failure
 * \sa SCE_Mem_Alloc SCE_Mem_Free
 * 
 * You will generally want to call SCE_calloc() that wraps this function.
 * 
 * \see SCE_calloc()
 */
void* SCE_Mem_Calloc (const char *file, unsigned int line, size_t s, size_t n)
{
    void *p = NULL;
    SCE_MemAlloc *mem = NULL;

    mem = SCE_Mem_NewAlloc ();

    p = calloc (s, n);
    if (!p)
    {
        SCE_Mem_DeleteAlloc (mem);
        Logger_Log (SCE_OUT_OF_MEMORY);
        return NULL;
    }

    mem->file = file;
    mem->line = line;
    mem->size = s * n;
    mem->ptr = p;

    SCE_Mem_AddAlloc (mem);

    return p;
}


/**
 * \brief Reallocs wrapper
 * \param aloc Old pointer
 * \param size New size wanted for the block
 * \param file, line Where the block is asked
 * \returns A pointer to the reallocated block on success, NULL on failure
 * \sa SCE_Mem_Alloc SCE_Mem_Free
 * 
 * You will generally want to call SCE_realloc() that wraps this function.
 * 
 * \see SCE_realloc()
 */
void* SCE_Mem_Realloc (const char *file, unsigned int line, void *p, size_t s)
{
    SCE_MemAlloc *mem = NULL;
    int addnew = 0;

    if (!p)
    {
        /* ajout d'une nouvelle allocation */
        addnew = 1;
        mem = SCE_Mem_NewAlloc ();
    }
    else
        mem = SCE_Mem_LocateAllocFromPointer (p);

    if (!mem)
    {
        Logger_Log ( (addnew) ? SCE_OUT_OF_MEMORY : SCE_INVALID_POINTER );
        return NULL;
    }

    p = realloc (p, s);
    if (!p)
    {
        if (addnew) /* en cas d'echec realloc conserve la memoire deja alloue */
            SCE_Mem_DeleteAlloc (mem);
        Logger_Log (SCE_OUT_OF_MEMORY);
        return NULL;
    }

    mem->file = file;
    mem->line = line;
    mem->size = s;
    mem->ptr = p;

    if (addnew)
        SCE_Mem_AddAlloc (mem);

    return p;
}

/* revise le 02/01/2008 */
/**
 * \brief Frees a pointer allocated by SCE_Mem_Alloc
 * 
 * You will generally want to call SCE_free() that wraps this function.
 * 
 * \see SCE_free
 */
void SCE_Mem_Free (void *p)
{
    if (p && SCE_Mem_RemoveAlloc (p) < 0)
        Logger_Log (SCE_INVALID_POINTER);
    else; /* this 'else' kick a warning message... wtf ? */
}


/* ajoute le 09/10/2007 */
void* SCE_Mem_Dup (void *p, size_t size)
{
    void *new = SCE_malloc (size);
    if (new)
        memcpy (new, p, size);
    else
        Logger_LogSrc ();
    return new;
}

/* ajoute le 15/10/2007 */
void SCE_Mem_Convert (int tdest, void *dest, int tsrc, void *src, size_t n)
{
    union SCE_UMemType
    {
        GLubyte *ub;
        GLbyte *b;
        GLushort *us;
        GLshort *s;
        GLuint *ui;
        GLint *i;
        GLfloat *f;
        GLdouble *d;
    };

    size_t i;
    union SCE_UMemType tin, tout;

#define SCE_MEM_FOR(type, namedest, namesrc)\
    case type:\
        tout.namedest = dest;\
        for (i=0; i<n; i++)\
            tout.namedest[i] = tin.namesrc[i];\
        break;

#define SCE_MEM_SWITCH(type, namesrc)\
    case type:\
        tin.namesrc = src;\
        switch (tdest)\
        {\
        SCE_MEM_FOR (SCE_UNSIGNED_BYTE,  ub, namesrc)\
        SCE_MEM_FOR (SCE_BYTE,           b,  namesrc)\
        SCE_MEM_FOR (SCE_UNSIGNED_SHORT, us, namesrc)\
        SCE_MEM_FOR (SCE_SHORT,          s,  namesrc)\
        SCE_MEM_FOR (SCE_UNSIGNED_INT,   ui, namesrc)\
        SCE_MEM_FOR (SCE_INT,            i,  namesrc)\
        SCE_MEM_FOR (SCE_FLOAT,          f,  namesrc)\
        SCE_MEM_FOR (SCE_DOUBLE,         d,  namesrc)\
        }\
        break;

    switch (tsrc)
    {
    SCE_MEM_SWITCH (SCE_UNSIGNED_BYTE,  ub)
    SCE_MEM_SWITCH (SCE_BYTE,           b)
    SCE_MEM_SWITCH (SCE_UNSIGNED_SHORT, us)
    SCE_MEM_SWITCH (SCE_SHORT,          s)
    SCE_MEM_SWITCH (SCE_UNSIGNED_INT,   ui)
    SCE_MEM_SWITCH (SCE_INT,            i)
    SCE_MEM_SWITCH (SCE_FLOAT,          f)
    SCE_MEM_SWITCH (SCE_DOUBLE,         d)
    }
}

/* ajoute le 15/10/2007 */
/* revise le 06/02/2008 */
void* SCE_Mem_ConvertDup (int tdest, int tsrc, void *src, size_t n)
{
    size_t size;
    void *dest = NULL;

    /* TODO: mosh */
    switch (tdest)
    {
    case SCE_DOUBLE:
        size = sizeof (GLdouble);
        break;

    case SCE_FLOAT:
        size = sizeof (GLfloat);
        break;

    case SCE_INT:
    case SCE_UNSIGNED_INT:
        size = sizeof (GLint);
        break;

    case SCE_SHORT:
    case SCE_UNSIGNED_SHORT:
        size = sizeof (GLshort);
        break;

    case SCE_BYTE:
    case SCE_UNSIGNED_BYTE:
        size = sizeof (GLbyte);

    default:
#ifdef SCE_DEBUG
        Logger_Log (SCE_INVALID_ARG);
        Logger_LogMsg ("invalid type for the destination");
#endif
        return NULL;
    }

    if (tdest == tsrc)
        return SCE_Mem_Dup (src, size * n);

    dest = SCE_malloc (size * n);
    if (!dest)
    {
        Logger_LogSrc ();
        return NULL;
    }

    SCE_Mem_Convert (tdest, dest, tsrc, src, n);
    return dest;
}

/* ajoute le 10/01/2008 */
void SCE_Mem_List (void)
{
    unsigned int n = 0;
    SCE_MemAlloc *a = NULL;
    SCE_Mem_For (a)
    {
        Logger_PrintMsg ("- allocation in %s (%u) : %u bytes.\n",
                         a->file, a->line, a->size);
        n++;
    }
    Logger_PrintMsg ("you have %u non-freeds allocations.\n", n);
}


#if 0
/**
 * \brief Get the file which allocated the pointer
 */
const char* SCE_Mem_GetFile (void *p)
{
    SCE_MemAlloc *m = SCE_Mem_LocateAllocFromPointer (p);
    if (m)
        return m->file;
    return NULL;
}

/**
 * \brief Get the line of the call to SCE_Mem_Alloc
 */
unsigned int SCE_Mem_GetLine (void *p)
{
    SCE_MemAlloc *m = SCE_Mem_LocateAllocFromPointer (p);
    if (m)
        return m->line;
    return 0;
}

/**
 * \brief Get the size used by a pointer
 */
size_t SCE_Mem_GetSize (void *p)
{
    SCE_MemAlloc *m = SCE_Mem_LocateAllocFromPointer (p);
    if (m)
        return m->size;
    return 0;
}
#endif

/** @} */
